//  -*- C++ -*-
//    File: ServoMon.hh
//
//    $Id: ServoMon.hh,v 1.8 2003/09/05 14:53:49 dwchin Exp $
#ifndef SERVOMON_HH
#define SERVOMON_HH

#include <iostream>
#include <map>
#include <string>
#include "OperStateCondList.hh"
#include "Interval.hh"
#include "TSWindow.hh"
#include "MonServer.hh"
#include "TrigRslt.hh"
#include "TrigClient.hh"
#include "DatEnv.hh"
#include "AlarmClient.hh"
#include "AlarmData.hh"
#include "Trend.hh"


//@{

/** @name ServoMon User's Guide

\begin{center}
{\bf{\Large Understanding the ServoMon DMT monitor}}
\end{center}

\begin{center}
{\bf{\large Abstract}}
\end{center}

This document gives an overview of ServoMon to help operators
interpret ServoMon's output and alarms.

\begin{center}
{\bf{\large Introduction}}
\end{center}

ServoMon monitors two channels for various conditions on their power
spectra (or, modulo a square root operation, band-limited RMS).  If 
certain conditions on their power spectra are met, alarms 
are set.

The channels that are monitored are:
\begin{itemize}
  \item {\tt XX:IOO-MC_F}
  \item {\tt XX:LSC-DARM_CTRL}
\end{itemize}
where {\tt XX} is one of {\tt H1}, {\tt H2}, {\tt L1}.

The frequency ranges that are monitored are:
\begin{itemize}
  \item {\tt IOO-MC_F}: 3000Hz - 7000Hz
  \item {\tt LSC-DARM_CTRL}: 340Hz - 345Hz (violin fundamental), 
    345Hz - 350Hz, 680Hz - 690Hz (violin 2nd harmonic), 400Hz - 450Hz
\end{itemize}

A summary status page is available via the web here 
{\tt http://blue.ligo-wa.caltech.edu/gds/monitor_reports/ServoMon/ServoMon_Status_XX.html}.  This page tabulates the problems that are being monitored, 
with the time of the last occurence, and the cumulative number of occurences 
since ServoMon began running.  All the web-page data is reset if ServoMon
is restarted.

ServoMon also serves time-series data that is available via the DMT
Viewer.  The time-series show the power for last 12 hours, similar to 
the BLRMS monitor plots.


*/

//@}



using namespace std;


//
//    Servo Instability monitor
//
class ServoMon
    : public DatEnv,
      public MonServer,
      private TrigClient
{
public:

    ServoMon(int argc, const char *argv[]);  // Constructor

    ~ServoMon();                             // Destructor

    void ProcessData();

    /** Interrupt for DMT Viewer
     */
    void Attention() { MonServer::Attention(); }

private:

    /// Name of monitor
    std::string       mName;

    void              appendToLog(Time now);

    void              updatePowerWindows(Time now);

    void              MCFproblemActions(Time now);
    void              D0problemActions(Time now);
    void              D0pProblemActions(Time now);
    void              D1problemActions(Time now);
    void              D2problemActions(Time now);

    void              statisticsActions(const Time & now);

    void              printCounters(void) const;

    size_t            MaxFrame;     //  Number of frames I should process
    size_t            mFrameCount;

    static const float mStride = 20.;

    /// The last frame no. when MC_F had problem
    int               mMCFLastProb;

    /// The last frame no. when DARM_CTRL had violin fund. problem
    int               mD0LastProb;

    /// The last frame no. when DARM_CTRL had 345-355Hz problem
    int               mD0pLastProb;

    /// The last frame no. when DARM_CTRL had violin 1st harm. problem
    int               mD1LastProb;

    /// The last frame no. when DARM_CTRL had 400-450Hz problem
    int               mD2LastProb;
    
    OperStateCondList mOsclist;
    
    /// Log file
    ofstream          mLogfile;     
    std::string       mLogfileName;
    ostream          *mLog;
    int               mDebug;

    /// Size of the TSWindows
    static const size_t mMaxLength = 2160;  // 12 hrs @ 20 sec. stride = 2160

    std::string mProblemPstr;
    std::string mMCFstr;
    std::string mD0str;
    std::string mD0plusstr;
    std::string mD1str;
    std::string mD2str;

    TSWindow<int>    mMCFProblemPWindow;
    TSWindow<double> mPowerWindowMCF;
    TSWindow<double> mPowerWindowD0;
    TSWindow<double> mPowerWindowD0plus;
    TSWindow<double> mPowerWindowD1;
    TSWindow<double> mPowerWindowD2;

    TSeries          mMCFProblemTS;
    TSeries          mPowerTSMCF;
    TSeries          mPowerTSD0;
    TSeries          mPowerTSD0plus;
    TSeries          mPowerTSD1;
    TSeries          mPowerTSD2;

    std::vector<double> mMCFdata;
    std::vector<double> mD0data;
    std::vector<double> mD0plusdata;
    std::vector<double> mD1data;
    std::vector<double> mD2data;

    std::string       mIFOname;

    std::string       mTrigType;
    std::string       mTrigSubtypeMCF;
    std::string       mTrigSubtypeD0;
    std::string       mTrigSubtypeD0plus;
    std::string       mTrigSubtypeD1;
    std::string       mTrigSubtypeD2;

    char              mOutputStr[256];

    //
    // Time stamps for when problems occur
    //
    char              mMCFtimestamp[64];
    char              mD0timestamp[64];
    char              mD0pTimestamp[64];
    char              mD1timestamp[64];
    char              mD2timestamp[64];

    //
    // counters to keep track of how many times problems
    // have occured
    //
    int               mMCFcounter;
    int               mD0counter;
    int               mD0pCounter;
    int               mD1counter;
    int               mD2counter;

    //
    // HTML status page stuff
    //
    /// Name of status display file
    std::string   mDispfileName;

    /// Livetime display html file
    std::ostream *mDisplay;

    /// Time stamp string
    char          mTimestamp[64];

    /// Time stamp format
    std::string   mTstampFormat;

    /// Update interval for HTML summary page (in seconds)
    static const int mHTMLupdateInterval = 300;

    /// Total time that monitor has been running
    Interval      mTotalRunTime;


    //
    // Alarm stuff
    //
    /// Alarm client
    AlarmClient       mAlarmClient;

    // Locations of files where description of alarms are contained.
    std::string       mMCF_alarmDesc;
    std::string       mD0_alarmDesc;
    std::string       mD0p_alarmDesc;
    std::string       mD1_alarmDesc;
    std::string       mD2_alarmDesc;

    // Each problem condition needs its own alarm data and handle
    AlarmData         mMCF_ad;
    AlarmData         mD0_ad;
    AlarmData         mD0p_ad;
    AlarmData         mD1_ad;
    AlarmData         mD2_ad;
    
    AlarmHandle       mMCF_ah;
    AlarmHandle       mD0_ah;
    AlarmHandle       mD0p_ah;
    AlarmHandle       mD1_ah;
    AlarmHandle       mD2_ah;

    /// HTML directory
    std::string       mHtmlDir;


    /// name of index page
    std::string       mIndexPageName;

    /// index page
    std::ofstream     mIndexPage;


    //
    // Trend stuff
    //

    /// Trend
    Trend             mTrend;

    /// Trend channel names
    map<std::string, std::string> mTrendChanNames;


    //
    // Trigger rate stuff
    //   now that TSWindows has a new feature to fillVector()
    //   we don't need one TSWindow for every rate computation
    //   I don't think
    // All of them will have a length of 4320 (24hrs @ 20secs)
    // GDS triggers to MetaDB are throttled to 1 per 5 min.
    // don't throttle these
    //
    TSWindow<int>    mMCFTriggers;
    TSWindow<int>    mD0Triggers;
    TSWindow<int>    mD0pTriggers;
    TSWindow<int>    mD1Triggers;
    TSWindow<int>    mD2Triggers;


    /// Print out HTML header
    void              htmlHead(void) const;

    /// Print out HTML footer
    void              htmlFoot(void) const;

    /// Print out HTML status page
    void              writeStatusPage(void);

    /// Print out index page
    void              writeIndexPage(void);

    /// Set TSWindows params
    void              setTSWindowStuff(void);

    /// Initialize TSWindows with dummy
    void              initTSWindows(const Time & now);

    /// Restore all TSWindows
    void              restoreAllTSWindows(void);

    /// Save all TSWindows
    void              saveAllTSWindows(void);

    /// Output table of trigger rates for status web page
    void makeTrigRateTable(const std::string & name,
                           double cumRate,
                           double twoHrRate,
                           double sixHrRate,
                           double twelveHrRate,
                           double oneDayRate);

    /** Trigger rate
      * @memo Returns trigger rate for the last (nhours, nminutes, nsecs)
      * @param nhours No. of hours in the past
      * @param nminutes No. of minutes in the past
      * @param nsecs No. of secs in the past
      */
    double triggerRate(const TSWindow<int> & triggers, size_t nhours, size_t nminutes=0, size_t nsecs=0);

    /// dummy vector for computing mean for trending
    std::vector<double>    mTmpVec;

};

#endif     //  SERVOMON_HH

