/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "mywv2lib.hh"
#include "kleineWelleN.hh"
#include "LPEFilter.hh"
#include "DVector.hh"

#include <iostream>
#include <iomanip>
#include <cstdlib>
#include <unistd.h>
#include <sstream>
#include <cmath>
#include <iterator>

#define KWVERSION "April 13, 2008"

EXECDAT(kleineWelle)

using namespace std;

//======================================  Channel data constructor
kwChan::kwChan(void)
  : filter("lpef"), significance(0.0), threshold(4.0), stride(16), fs(0), 
    size(0), lpefLength(0), maxDist(2), scale(0), lowFrequency(64), 
    highFrequency(1024), lowScale(0), highScale(0), decimateFactor(-1), 
    transientDuration(4.0), wavePtr(NULL), prefilter(NULL), lpefilter(NULL), 
    nextTime(0), factor(1.0), waveInjection(NULL)
{
}

//======================================  Channel data constructor
kwChan&
kwChan::operator=(const kwChan& x) {
    channel        = x.channel;
    filter         = x.filter;
    filename       = x.filename;
    significance   = x.significance;
    threshold      = x.threshold;
    stride         = x.stride;
    fs             = x.fs;
    size           = x.size;
    lpefLength     = x.lpefLength;
    maxDist        = x.maxDist;
    scale          = x.scale;
    lowFrequency   = x.lowFrequency;
    highFrequency  = x.highFrequency;
    lowScale       = x.lowScale;
    highScale      = x.highScale;
    decimateFactor = x.decimateFactor; 
    transientDuration = x.transientDuration;
    wavePtr        = NULL;
    if (x.prefilter) prefilter  = x.prefilter->clone();
    else             prefilter  = NULL;
    if (x.lpefilter) lpefilter  = x.lpefilter->clone();
    else             lpefilter  = NULL;
    nextTime       = x.nextTime;
    injections     = x.injections;
    factor         = x.factor;
    waveInjection  = NULL; // cheat rather than allocating/copying a new string
    return *this;
}

//======================================  Channel destructor
kwChan::~kwChan(void) {
    if(wavePtr) {
	delete wavePtr;
	wavePtr   = NULL;
    } 
    if(prefilter) {
	delete prefilter;
	prefilter = NULL;
    }
    if(lpefilter) {
	delete lpefilter;
	lpefilter = NULL;
    }
}
    
//======================================  Print out channel parameters
std::ostream& 
kwChan::dump_params(std::ostream& out) const {
    out << "# Analyzing channel: " << channel << endl;
    if(!injections.empty()) {
	out << "# with injections: " << injections << endl;
	out << "# with multiplication factor: " << factor << endl;
    }
    out << "# with transient duration: " << transientDuration << endl;
    out << "# with black pixel threshold: " << threshold << endl;
    out << "# with cluster signif threshold: " << significance << endl;
    out << "# with maxDist between black pix: " << maxDist << endl;
    return out;
}

void 
kwChan::setup(void) {
    // need to cap high frequency boundary to nyquist
    if(highFrequency > fs / 2)
	highFrequency = fs /2;
    if(highScale)
	lowFrequency = fs / (1 << highScale);
    else
	highScale = (int)ceil(log2(fs / lowFrequency));
    if(lowScale)
	highFrequency = fs / (1 << lowScale);
    else {
	lowScale = (int)floor(log2(fs / highFrequency));
    }
    cout << "# over approximate frequency range " << lowFrequency 
	 << " - " << highFrequency << endl;
    cout << "# over wavelet scales " << lowScale << " to " 
	 << highScale << endl;
    if(decimateFactor == -1) {   // decimate factor is not set in options
	decimateFactor = lowScale - 1;  // leave headroom for linear filter
    }
    cout << "# decimating factor " << decimateFactor << endl;
    FilterDesign fd(fs);
    fs = (fs >> decimateFactor);
    cout << "# decimating down to " << fs << endl;
    lowScale -= decimateFactor;
    highScale -= decimateFactor; 
    size = long(fs * double(stride) + 1e-6);
    wavePtr = new Haar_Wavelet(size);
    fd.decimateBy2(decimateFactor, 3); // type 3 has most ringing, but least timeshift
    //------------------------------  With filtering defined add prefilters
    //                                to fd filterdesign
    if(!filter.empty()) {
	if(filter == "hpf" || filter == "lpef") {
	    fd.butter(kHighPass, 6, lowFrequency, 0);
	    cout << "# using high pass filtering at " << lowFrequency 
		 << endl;
	    if(filter == "lpef" && !lpefLength) {
		// assume haar Q =~ 4, this provides sufficient whitening
		// actual form should be order > 'Q * fs / minFreq' as 
		// df = freq / Q  and  df = fs / order
		// however do, not exceed sample rate for filter order 
		// because we only have 1sec to account for ringing 
		// (still not enough because of HPF ringing)
		lpefLength = min(fs, int(16*highFrequency/lowFrequency));
	    }
	}
	else {
	    fd.filter(filter.c_str());  // add custom filter
	    cout << "# using prefilter " << filter << endl;
	}
	if(lpefLength) {
	    lpefilter = new LPEFilter(lpefLength, 0, int(double(stride)*fs)-1);
	    cout << "# using linear predictive filtering of order " 
		 << lpefLength << endl;
	}
    }
    // grab prefilter pipe, will contain decimator and possible hpf
    prefilter = fd.release();
    nextTime = Time(0);
    cout << "# output: start, stop, central time, frequency, energy, "
	 << "normalized energy, number of pixels, significance" << endl;

    //------------------------------------ read injection waveform from file
    if(!filename.empty()) {
	ifstream waveFile(filename.c_str(), ios::in);
	if(!waveFile.is_open()) {
	    cerr << "cannot open waveform file: " << filename << endl;
	    cerr << "INJECTION IGNORED!" << endl;
	} else {
	    string waveString;
	    float sample;
	    float hrss = 0;
	    double hss = 0;
	    int begin = size / 2;
	    waveInjection = new float[size];
	    for(int i = 0; i < size; i++) {
		waveInjection[i] = 0;
	    }

	    //---------------------------  fill second half of injection 
	    //                             data with waveform
	    int length = 0;
	    for(int i=0; i < begin; i++) {
		getline(waveFile, waveString);
		if(waveFile.eof()) {
		    break;
		}
		istringstream waveStream(waveString);
		waveStream >> sample;
		waveInjection[begin + i] = sample * factor;
		hss += sample * sample;
		length++;
	    }
	    hrss  = hss * factor / fs;
	    cout << "# read in waveform injection of length " << length
		 << " and scaled hrss of " << hrss << endl;
	}
    }
}

//======================================  Channel processing method
void
kwChan::process(MultiDacc& dacc) {
    //----------------------------------  Set up local storage
    //  lsize -     start index of the wavelet coefficients in a scale
    //  tsIn -      raw channel time series. converted to float.
    //  t0 -        Start time of stride.
    TSeries* tsIn = dacc.refData(channel.c_str());
    tsIn->Convert(DVector::t_float);
    Time t0 = tsIn->getStartTime();

    //----------------------------------  Scale injection and add them to tsIn
    if(!injections.empty()) {
	*(dacc.refData(injections.c_str())) *= factor;
	*tsIn += *(dacc.refData(injections.c_str()));
    }

    //------------------------------- set sample rate and filters if not defined
    if(!fs) {
	fs = int(tsIn->getNSample() / tsIn->getInterval() + 1e-6);
	printf("# detected sampling rate %d\n", fs);
	dump_params(cout);
	setup();
    }

    //----------------------------------  Append input series to input buffer
    //
    //  If the input is not continunous with the previous stride, reset the
    //  pre-filters and set the discard the start of the data.
    //
    if(t0 != nextTime) {
	if (nextTime != Time(0)) {
	    cout << "# detected break in data: expected " << setprecision(12)
		 << nextTime.totalS() << " got " << t0.totalS() << endl;
	}
	prefilter->reset();
	inputBuffer.Clear();
	inputBuffer = prefilter->apply(*tsIn);
	if(lpefLength) {
	    lpefilter->setHistory(inputBuffer.extract(t0,transientDuration));
	}
	inputBuffer.eraseStart(transientDuration);
    } 

    //-----------------------------------  prefilter the stride into the buffer
    else {

	//-------------------------------  add possible waveform injection
	if(waveInjection) {
	    *tsIn += TSeries(tsIn->getStartTime(), tsIn->getTStep(), 
			     size, waveInjection);
	    cout << "# added injection waveform vector" << endl;
	}

	inputBuffer.Append(prefilter->apply(*tsIn));
    }

    //-----------------------------------  Continuous data, run the KW pipeline
    while (inputBuffer.getInterval() >= stride) {

	//-------------------------------  Perform LPE filtering
	Interval delay = prefilter->getTimeDelay();
	Time tStart = inputBuffer.getStartTime();
	TSeries strideData;
	if(lpefLength) {
	    TSeries temp = inputBuffer.extract(tStart, stride);
	    lpefilter->train(temp);
	    lpefilter->apply(temp, strideData);
	    delay += lpefilter->getTimeDelay();
	} else {
	    strideData = inputBuffer.extract(tStart, stride);
	}

	//------------------------------  Do the wavelet transform...
	//                                put data into the wavelet class.
	strideData.getData(size, wavePtr->getRawArray());
	printf("# processing at t0 = %f\n", tStart.totalS());

	//------------------------------  Do transform, Get unnormalized coefs
	wavePtr->transform(1);
	double* wavearray = wavePtr->getWaveArray();

	//------------------------------- create vector of pixels over threshold
	element_vect elements;
	for(int scale = lowScale; scale <= highScale; scale++) {

	    //---------------------------  wavelet width, number of elements
	    double dtScale = double(1 << scale) / double(fs);
	    int      lsize = size / (1 << scale);

	    //---------------------------  Get wavelet coef statistics. Note 
	    //                             that this is moved from the original
	    double lsum = 0;
	    for(int i=lsize; i<2*lsize; i++) {
		lsum += wavearray[i]*wavearray[i];
	    }
	    float sigi = sqrt(lsum / lsize);  // float!!!

	    //---------------------------  look for black pixels.
	    for(int i = lsize; i < 2*lsize; i++) {
		double wi = wavearray[i];
		if(abs(wi / sigi) > threshold) {
		    //1/double time = tStart.getS() + 1e-9 * tStart.getN() 
		    //1/            + dtScale * (i-lsize);
		    double time = dtScale * double(i-lsize);
		    //1/-----------------------------------------------------
		    //---------------------- linear phase delay from filters
		    //                       can be removed here!!! I'll leave
		    //                       this out until after testing and 
		    //                       debugging.
		    // time -= delay;
		    Element el(wi*wi, pow(wi/sigi, 2), i, scale, time, 
			       time + dtScale, dtScale, 0);
		    elements.push_back(el);
		}
	    }
	}

	//-----------------------------  Assign clusterID to each black pixel
	int freeID = 0;
	int N = elements.size();
	if (N > 0) {
	    vector<int> pvect(N, 0);
	    for (int i=0; i<N; ++i) pvect[i] = i;

	    int ifree = 0;
	    for (int i=0; i<N; ++i) {
		if (i == ifree) {
		    ifree++;
		    elements[pvect[i]].clusterID = freeID++;
		}
		const Element& testEl(elements[pvect[i]]);
		for (int j=ifree; j<N; ++j) {
		    int pj = pvect[j];
		    if (testEl.dist(elements[pj]) <= maxDist) {
			elements[pj].clusterID = testEl.clusterID;
			if (j != ifree) {
			    pvect[j]     = pvect[ifree];
			    pvect[ifree] = pj;
			}
			ifree++;
		    }
		}
	    }
	}

	//------------------------------------- sum clusters by ID
	event_vect events(freeID, Event(channel));
	for(element_iter i=elements.begin(); i < elements.end(); i++) {
	    events[i->clusterID] += *i;
	}

	//--------------------------------------- estimate cluster parameters
	int j=0;
	for(int i = 0; i < freeID; i++) {
	    events[i].average(tStart, fs);
	    if(events[i].significance >= significance) { 
		if (j != i) events[j] = events[i];
		j++;
	    }
	}

	//------------------------------  output sorted events
	if (j) {
	    events.erase(events.begin()+j, events.end());
	    sort(events.begin(), events.end());
	    copy(events.begin(), events.end(), ostream_iterator<Event>(cout));
	}

	//------------------------------  Delete data at start of input buffer.
	inputBuffer.eraseStart(stride);
    }
    nextTime = tsIn->getEndTime();
}

//======================================  Event null constructor
Event::Event(const std::string& name)
    : startTime(0), endTime(0), centralTime(0), frequency(0), 
      energy(0), normEnergy(0), npts(0), significance(0), 
      channelName(name) 
{}

//======================================  Add an element
Event& 
Event::operator += (const Element& el) {
    if (npts++) {
	startTime = min(startTime, el.startTime);
	endTime   = max(endTime, el.endTime);
    } else {
	startTime = el.startTime;
	endTime   = el.endTime;
    }
    double norm = el.normEnergy;
    centralTime += norm * (el.startTime + el.endTime)/2.;
    frequency += el.scale * norm;
    energy += el.energy;
    normEnergy += norm;
    return *this;
}

//======================================  Event printout operator.
void 
Event::average(const Time& toff, double fs) {
    if (!npts) return;
    //1/centralTime = centralTime / normEnergy;
    centralTime = toff.totalS() + centralTime / normEnergy;
    startTime += toff.totalS();
    endTime   += toff.totalS();
    //1/----------------------------------------
    significance = logSignificance(normEnergy/2.0, npts/2.0, .01);
    frequency = fs / pow(2.0, frequency / normEnergy);
}

//======================================  Event printout operator.
std::ostream& 
operator<<(std::ostream& output, const Event& a) {
    char line[128];
    sprintf(line, "%16.6f  %16.6f  %16.6f  %4.0f  %9.3e  %6.2f  %3d  %6.2f %s",
	    a.startTime, a.endTime, a.centralTime, a.frequency, a.energy, 
	    a.normEnergy, a.npts, a.significance, a.channelName.c_str());
    output << line << endl;
    return output;
}

kleineWelle::kleineWelle(int argc, const char *argv[])
    : MultiStream(argc, argv), segmentStart(0), segmentStop(0), stride(16.0)
{
    bool syntax = false;    // flag for syntax error

    //----------------------------------  parse command syntax
    std::ifstream optionsFile; // file stream used only in initialization.
    if(argc == 1) {
	syntax = true;
    } else {
	optionsFile.open(argv[1], ios::in);
	if(!(optionsFile.is_open())) {
	    cerr << "cannot open options file: " << argv[1] << endl;
	    finish();
	    return;
	} else if (Debug()) {
	    cerr << "reading options file: " << argv[1] << endl;
	}
    }
    if(syntax) {
	cerr << "kleineWelle version " << KWVERSION << endl;
	cerr << "Command syntax:" << endl;
	cerr << argv[0] << " [<options file>] [<MultiStream-args>]" << endl;
	cerr << "where the options file is a two column ascii file which defines "
	     << "the following options:" << endl;
	cerr << "stride, transientDuration, channel, injections, factor, filter, "
	     << "threshold, maxDist, segmentFile, lowFrequency, highFrequency" 
	     << endl;
	cerr << "please see sample options file 'opt' for description of these "
	     << "parameters" << endl;
	finish(); 
	return;
    }

    //-------------------------------------- read in options from file
    string segment_file_name;
    kwChan default_chan, temp;
    while(!optionsFile.eof()) {
	string option, word;
	getline(optionsFile, option);
	if (Debug()) cerr << "Option: " << option << endl;
	istringstream optionsStream(option);
	optionsStream >> word;
	if(word.length() == 0 || word[0] == '#') {
	    if (Debug()) cerr << "Option skipped" << endl;
	    continue;
	}
	else if(word == "stride") {         // DMT analysis stride
	    double epoch;
	    optionsStream >> epoch;
	    stride = Interval(epoch);
	}
	else if(word == "channel_stride") { // Channel analysis stride
	    double epoch;
	    optionsStream >> epoch;
	    temp.stride = Interval(epoch);
	}
	else if(word == "channel") {         // KW analysis channel
	    if (temp.channel.empty() || temp.channel == "default") {
		default_chan = temp;
	    } else {
		if (Debug()) cerr << "Configure channel: " << temp.channel 
				  << endl;
		mChanList.push_back(temp);
		temp = default_chan;
	    }
	    optionsStream >> temp.channel;
	}
	else if(word == "injections")       // KW software injections channel
	    optionsStream >> temp.injections;
	else if(word == "factor")           // injections scale factor
	    optionsStream >> temp.factor;
	else if(word == "filter")           // prefilter type
	    optionsStream >> temp.filter;
	else if(word == "threshold")        // threshold on coefficient amplitude
	    optionsStream >> temp.threshold;
	else if(word == "significance")     // threshold on cluster significance
	    optionsStream >> temp.significance;
	else if(word == "maxDist")          // distance for clustering
	    optionsStream >> temp.maxDist;
	else if(word == "segmentFile") {    // segments to analyze
	    optionsStream >> segment_file_name;
	}
	else if(word == "waveFile") {        // ASCII file injection waveform
	    optionsStream >> temp.filename;
	}
	//------ most of these options are automatically calculated but take precedence if defined
	else if(word == "lpefLength")     // obsoleted (automatically decided)
	    optionsStream >> temp.lpefLength;
	else if(word == "lpefOrder")      // obsoleted
	    optionsStream >> temp.lpefLength;
	else if(word == "lowScale")       // obsoleted but take precedence if defined
	    optionsStream >> temp.lowScale;
	else if(word == "highScale")      // obsoleted but take precedence if defined
	    optionsStream >> temp.highScale;
	else if(word == "lowFrequency")   // low frequency of analysis (sets highScale)
	    optionsStream >> temp.lowFrequency;
	else if(word == "highFrequency")  // high frequency of analysis (sets lowScale)
	    optionsStream >> temp.highFrequency;
	else if(word == "decimateFactor") // # times to decimate data(automatic)
	    optionsStream >> temp.decimateFactor;
	else if(word == "transientDuration") { // length of filter transient/
	    double epoch;
	    optionsStream >> epoch;
	    temp.transientDuration = Interval(epoch);
	}
	else
	    cerr << "bad option: " << word << endl;
    }
    if (!temp.channel.empty()) {
	if (Debug()) cerr << "Configure channel: " << temp.channel << endl;
	mChanList.push_back(temp);
    }
    optionsFile.close();

    //----------------------------------  Check at least one channel
    if (mChanList.empty()) {
	cerr << "No channels were configured!" << endl;
	finish(); 
	return;
    }

    //----------------------------------  Print out configured parameters
    cout << "# kleineWelle triggers" << endl;
    cout << "# Algorithm version: " KWVERSION << endl;
    cout << "# with stride: " << stride << endl;

    //----------------------------------  Set up data stream.
    getDacc().setStride(stride);

    //----------------------------------  Check the number of streams
    int nStream = getNStream();
    if (!nStream || nStream > 2) {
	cerr << "kleineWelle currently supports 0 < n <= 2 streams only" 
	     << endl;
	finish(); 
	return;
    }
 
    // set noise channel in stream 1
    // "sim" data needs "addSimulated", "adc" and "proc" data needs "addChannel"
    // set injections channel to stream 1 or 2 depending on number of streams
    for (channel_iter i=mChanList.begin(); i != mChanList.end(); ++i) {
	getDacc().addChannel(i->channel.c_str(), 0);
	if(!i->injections.empty()) {
	    getDacc().addChannel(i->injections.c_str(), nStream-1);
	}
    }

    //----------------------------------  Read in segment data
    if (segment_file_name.empty()) {
	segmentStart = Time(0);
	segmentStop  = Time(0);
    } else {
	ifstream segmentFile(segment_file_name.c_str());
	if(!segmentFile.is_open()) {
	    cerr << "Unable to open segment file: " << segment_file_name 
		 << endl;
	    finish();
	    return;
	} else {
	    string segmentString;
	    while (!segmentFile.eof()) {
		getline(segmentFile, segmentString); // get next segment or EOF
		if (segmentString.empty()) continue;
		if (segmentString[0] == '#') continue;
		istringstream segmentStream(segmentString);
		gps_type start, stop;
		segmentStream >> start >> stop;
		mSegVect.push_back(segment_node(start, stop));
	    }
	    pop_segment();
	}
    }
}

kleineWelle::~kleineWelle()
{
    cout << "# finished." << endl;
}

//======================================  Pop a new segment off the list.
bool
kleineWelle::pop_segment(void) {
    if (mSegVect.empty()) return false;
    segmentStart = Time(mSegVect[0].first);
    segmentStop  = Time(mSegVect[0].second);
    mSegVect.erase(mSegVect.begin());
    cout << "# Analyzing segment " << segmentStart.getS() << " to " 
	 << segmentStop.getS() << endl;
    if (getDacc().getCurrentTime() < segmentStart) {
	getDacc().seek(segmentStart);
    }
    return true;
}

//======================================  Process one data stride
void kleineWelle::ProcessData(void) {

    //----------------------------------  Make sure were still inside the 
    //                                    segment... Loop over channels.
    if (!segmentStop || getDacc().getCurrentTime() < segmentStop) {
	for (channel_iter i=mChanList.begin(); i != mChanList.end(); ++i) {
	    i->process(getDacc());
	}
    }

    //----------------------------------  Anticipate end of data stream as 
    //                                    we approach segment end the next 
    //                                    ProcessData() will grab the next 
    //                                    'stride' of data
    if(segmentStop != Time(0)) {
	Time t0 = getDacc().getDacc(0)->getFillTime();
	Time tEnd = t0 + 2.0 * stride;
	if(tEnd > segmentStop) { // if the next data goes past masktop
	    while(tEnd > Time(segmentStop)) {
		if (!pop_segment()) {
		    finish();
		    break;
		}
	    }
	}
    }
}

//---------------------------------------  negative natural log of significance
double
logSignificance(double normalizedEnergy, double timeFrequencyArea, 
		double precision)
{
    /*
     * This function assumes time-frequency measurement of stationary white
     * Gaussian noise.  The resulting measurement (2*normalizedEnergy) are 
     * chi squared distributed with 2A degrees of freedom, where A is the 
     * time-frequency area of the measurement.  This function returns the 
     * negative natural logarithm of the probability of achieving a less 
     * likely result.  It uses a series expansion of the regularized upper 
     * incomplete gamma function which is valid for the tail (highly 
     * significant) portion of the distribution.  See Ambramowitz and Stegun, 
     * Handbook of Mathematical Functions, section 6.1, page 264.
     */

    // maximum number of terms in series expansion
    int maximumTerms = 10;

    // initialize series expansion
    double sum = 1;
    double numerator = 1;
    double denominator = 1;
    double term;

    // compute and sum series terms
    for (term = 1; term <= maximumTerms; term++) {  
	numerator *= (timeFrequencyArea - term) * (timeFrequencyArea > term);
	denominator *= normalizedEnergy;
	double ratio = numerator / denominator;
	sum += ratio;
	if (ratio < precision * sum) {
	    break;
	}
    }

    // give error if maximum number of terms was exceeded
    if (term > maximumTerms) {
	std::cerr << "logSignificance: maximum number of terms exceeded"
		  << std::endl;
    }

    // negative natural logarithm of significance
    double logSignificance = normalizedEnergy -
	(timeFrequencyArea - 1) * log(normalizedEnergy) -
	log(sum) + lgamma(timeFrequencyArea);
  
    // return to calling function
    return logSignificance;
  
}
