#ifndef AUTOPIPE_HH
#define AUTOPIPE_HH

#include "Pipe.hh"

/**  auto_pipe maintains a pipe pointer and allows easy access to the 
  *  basic functionality of the Pipe class. If the Pipe class pointer 
  *  is null the pipe functions behave as a unit pipe, i.e. ts == unit(ts).
  *  @memo Automatic pipe pointer class.
  *  @version 1.0; Last modified June 5, 2006.
  *  @author J. Zweizig
  */
class auto_pipe {
public:
  /**  Construct a null pipe.
   */
  auto_pipe(void);

  /**  Construct an auto_pipe with a clone of the argument pipe.
   */
  auto_pipe(const Pipe& p);

  /**  Construct a pipe with a clone of the argument auto_pipe.
   */
  auto_pipe(const auto_pipe& p);

  /**  Construct an auto_pipe with a clone using the specified pipe.
   */
  auto_pipe(Pipe* p);

  /**  Destroy an auto_pipe.
   */
  virtual ~auto_pipe(void);

  /**  Copy a clone of the argument auto_pipe.
    *  @memo Assignment operator.
    *  @param ts TSeries to be filtered.
    *  @return reference to to this object
    */
  auto_pipe& operator=(const auto_pipe& ap);

  /**  Filter the argument time series using the referenced pipe. If the 
    *  pipe pointer is NULL the returned TSeries is identical to the 
    *  argument time series.
    *  @memo Filter a TSeries
    *  @param ts TSeries to be filtered.
    *  @return Filtered time series.
    */
  TSeries operator()(const TSeries& ts);

  /**  Dereference the Pipe pointer.
    */
  Pipe* operator->(void);
  Pipe& operator*(void);
  const Pipe* operator->(void) const;
  const Pipe& operator*(void) const;

  /**  Run the dataCheck method of the referenced Pipe on the argument 
    *  time series. No operation is performed if the pointer is null.
    *  @memo Pipe input data check.
    *  @param ts Time series to be checked.
    */
  void dataCheck(const TSeries& ts) const;

  /**  Run the isDataValid method of the referenced Pipe on the argument 
    *  time series. isDataValid returns true if the pointer is null.
    *  @memo Pipe input data check.
    *  @param ts Time series to be checked.
    */
  bool isDataValid(const TSeries& ts) const;

  /**  Test for a Null pipe pointer.
    *  @memo Test for null pipe.
    */
  bool null(void) const;

  /**  Reset the referenced Pipe using the Pipe::reset method. If the 
    *  Pipe pointer is null no operation is performed.
    */
  void reset(void);

  /**  Release the referenced pipe.
    */
  Pipe* release(void);

  /**  Set the Pipe pointer to a clone of the argument Pipe.
   */
  void set(const Pipe& p);

  /**  Set the #Pipe# pointer to the argument object. The #auto_pipe# class
    *  acquires ownership of the argument Pipe object.
    */
  void set(Pipe* p);

private:
  Pipe* mPipe;
};

//======================================  Inline definitions 
inline
auto_pipe::auto_pipe(void)
  : mPipe(0)
{}

inline void 
auto_pipe::set(Pipe* p) {
    if (mPipe) delete mPipe;
    mPipe = p;
}

inline bool 
auto_pipe::null(void) const {
    return !mPipe;
}

inline
auto_pipe::auto_pipe(Pipe* p) 
  : mPipe(p)
{}

inline void 
auto_pipe::set(const Pipe& p) {
    set(p.clone());
}

inline
auto_pipe::auto_pipe(const Pipe& p)
  : mPipe(0)
{
    set(p);
}

inline const Pipe&
auto_pipe::operator*(void) const {
    return *mPipe;
}

inline Pipe&
auto_pipe::operator*(void) {
    return *mPipe;
}

inline auto_pipe&
auto_pipe::operator=(const auto_pipe& ap) {
    if (ap.null()) set(0);
    else           set(*ap);
    return *this;
}

inline
auto_pipe::auto_pipe(const auto_pipe& ap)
  : mPipe(0) 
{
    *this = ap;
}

inline
auto_pipe::~auto_pipe(void) {
    set(0);
}

inline TSeries 
auto_pipe::operator()(const TSeries& ts) {
    if (null()) return ts;
    return mPipe->operator()(ts);
}

inline Pipe*
auto_pipe::operator->(void) {
    return mPipe;
}

inline const Pipe*
auto_pipe::operator->(void) const {
    return mPipe;
}

inline void 
auto_pipe::dataCheck(const TSeries& ts) const {
    if (!null()) mPipe->dataCheck(ts);
}

inline bool
auto_pipe::isDataValid(const TSeries& ts) const {
    if (null()) return true;
    return mPipe->isDataValid(ts);
}

inline void
auto_pipe::reset(void) {
    if (!null()) mPipe->reset();
}

inline Pipe* 
auto_pipe::release(void) {
    Pipe* p = mPipe;
    mPipe = 0;
    return p;
}

#endif // AUTOPIPE_HH
